/*
    This program generates meshes for tetrahedral elements.  It does it by
    sub-division.  It is taken from the Matlab functions tetra4_sphere.m
    and tetra4_cube.m in the package FSELIB for chapter 6 written by C. Pozrikidis
    from his book:

 Pozrikidis, C., Introduction to Finite and Spectral Element Methods using Matlab,
      Chapman & Hall/CRC, Boca Raton, 2005.

    You can get the complete FSELIB package at:

      http://dehesa.freeshell.org/FSELIB/

                Updated 1/9/07 Translated into C by San Le

                Copyright C. Pozrikidis

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define SMALL   0.000001

int ne, ng, ndiv;
int numnp, numel, shape_flag, triangle_flag = 0, circle_flag;
int npel = 4, nsd = 3;

int tetrareader(int *, char *, int , FILE *, double *, double *, double *);

int slffeawrite(double *, int *, int, int);

int tetra4( double *x, double *y, double *z, double *p, int *efl, int *c, int ndiv)
{
	int i,i2,j,k,dum,dum2,dum3,dum4;
	double fdum, fdum1, fdum2, fdum3, fdum4, Iflag;
	double *xn, *yn, *zn, xp[40], yp[40], rad;
	double xx[4], x12, x13, x14, x23, x24, x34;
	double yy[4], y12, y13, y14, y23, y24, y34;
	double zz[4], z12, z13, z14, z23, z24, z34;
	double efl12, efl13, efl14, efl23, efl24, efl34;
	int *efln, *gfl;

	int nm;

	xn = (double *)calloc(numel*8*ndiv*4,sizeof(double));
	yn = (double *)calloc(numel*8*ndiv*4,sizeof(double));
	zn = (double *)calloc(numel*8*ndiv*4,sizeof(double));
	efln = (int *)calloc(numel*8*ndiv*4,sizeof(int));
	gfl = (int *)calloc(numnp*3,sizeof(int));

/*
  refinement loop
*/


	if(ndiv > 0)
	{
	    for( i = 0; i < ndiv; ++i )
	    {
		nm = 0; /* count the new elements arising by each refinement loop
			   eight elements will be generated in each pass */

		for( j = 0; j < ne; ++j )   /* loop over current elements */
		{

		    for( k = 0; k < 4; ++k )
		    {
			xx[k]=x[4*j+k]; yy[k]=y[4*j+k]; zz[j]=z[4*j+k];
		    }

		    /* mid-edge nodes */

		    x12=0.5*(x[4*j+0]+x[4*j+1]); y12=0.5*(y[4*j+0]+y[4*j+1]);
		    z12=0.5*(z[4*j+0]+z[4*j+1]);
		    efl12 = 0; if(efl[4*j+0]==1 & efl[4*j+1]==1) efl12 = 1;
		    x13=0.5*(x[4*j+0]+x[4*j+2]); y13=0.5*(y[4*j+0]+y[4*j+2]);
		    z13=0.5*(z[4*j+0]+z[4*j+2]);
		    efl13 = 0; if(efl[4*j+0]==1 & efl[4*j+2]==1) efl13 = 1;
		    x14=0.5*(x[4*j+0]+x[4*j+3]); y14=0.5*(y[4*j+0]+y[4*j+3]);
		    z14=0.5*(z[4*j+0]+z[4*j+3]);
		    efl14 = 0; if(efl[4*j+0]==1 & efl[4*j+3]==1) efl14 = 1;
		    x23=0.5*(x[4*j+1]+x[4*j+2]); y23=0.5*(y[4*j+1]+y[4*j+2]);
		    z23=0.5*(z[4*j+1]+z[4*j+2]);
		    efl23 = 0; if(efl[4*j+1]==1 & efl[4*j+2]==1) efl23 = 1;
		    x24=0.5*(x[4*j+1]+x[4*j+3]); y24=0.5*(y[4*j+1]+y[4*j+3]);
		    z24=0.5*(z[4*j+1]+z[4*j+3]);
		    efl24 = 0; if(efl[4*j+1]==1 & efl[4*j+3]==1) efl24 = 1;
		    x34=0.5*(x[4*j+2]+x[4*j+3]); y34=0.5*(y[4*j+2]+y[4*j+3]);
		    z34=0.5*(z[4*j+2]+z[4*j+3]);
		    efl34 = 0; if(efl[4*j+2]==1 & efl[4*j+3]==1) efl34 = 1;

/* assign vertex nodes to sub-elements */
/* these will become the "new" elements */

		    xn[4*nm+0]=x13;    yn[4*nm+0]=y13;    zn[4*nm+0]=z13;    efln[4*nm+0]=efl13;
		    xn[4*nm+1]=x23;    yn[4*nm+1]=y23;    zn[4*nm+1]=z23;    efln[4*nm+1]=efl23;
		    xn[4*nm+2]=x[4*j+2]; yn[4*nm+2]=y[4*j+2]; zn[4*nm+2]=z[4*j+2]; efln[4*nm+2]=efl[4*j+2];
		    xn[4*nm+3]=x34;    yn[4*nm+3]=y34;    zn[4*nm+3]=z34;    efln[4*nm+3]=efl34;
		    /*printf( "aaa %4d %4d %16.8e %16.8e %16.8e %16.8e\n", 1, nm,  zn[4*nm+0], zn[4*nm+1], zn[4*nm+2], zn[4*nm+3]);*/
		    nm += 1;  /* first subelement */

		    xn[4*nm+0]=x12;    yn[4*nm+0]=y12;    zn[4*nm+0]=z12;    efln[4*nm+0]=efl12;
		    xn[4*nm+1]=x[4*j+1]; yn[4*nm+1]=y[4*j+1]; zn[4*nm+1]=z[4*j+1]; efln[4*nm+1]=efl[4*j+1];
		    xn[4*nm+2]=x23;    yn[4*nm+2]=y23;    zn[4*nm+2]=z23;    efln[4*nm+2]=efl23;
		    xn[4*nm+3]=x24;    yn[4*nm+3]=y24;    zn[4*nm+3]=z24;    efln[4*nm+3]=efl24;
		    nm += 1;  /* second subelement */

		    xn[4*nm+0]=x14;    yn[4*nm+0]=y14;    zn[4*nm+0]=z14;    efln[4*nm+0]=efl14;
		    xn[4*nm+1]=x24;    yn[4*nm+1]=y24;    zn[4*nm+1]=z24;    efln[4*nm+1]=efl24;
		    xn[4*nm+2]=x34;    yn[4*nm+2]=y34;    zn[4*nm+2]=z34;    efln[4*nm+2]=efl34;
		    xn[4*nm+3]=x[4*j+3]; yn[4*nm+3]=y[4*j+3]; zn[4*nm+3]=z[4*j+3]; efln[4*nm+3]=efl[4*j+3];
		    nm += 1;  /* third subelement */

		    xn[4*nm+0]=x[4*j+0]; yn[4*nm+0]=y[4*j+0]; zn[4*nm+0]=z[4*j+0]; efln[4*nm+0]=efl[4*j+0];
		    xn[4*nm+1]=x12;    yn[4*nm+1]=y12;    zn[4*nm+1]=z12;    efln[4*nm+1]=efl12;
		    xn[4*nm+2]=x13;    yn[4*nm+2]=y13;    zn[4*nm+2]=z13;    efln[4*nm+2]=efl13;
		    xn[4*nm+3]=x14;    yn[4*nm+3]=y14;    zn[4*nm+3]=z14;    efln[4*nm+3]=efl14;
		    nm += 1;  /* fourth subelement */

		    xn[4*nm+0]=x14; yn[4*nm+0]=y14; zn[4*nm+0]=z14; efln[4*nm+0]=efl14;
		    xn[4*nm+1]=x13; yn[4*nm+1]=y13; zn[4*nm+1]=z13; efln[4*nm+1]=efl13;
		    xn[4*nm+2]=x34; yn[4*nm+2]=y34; zn[4*nm+2]=z34; efln[4*nm+2]=efl34;
		    xn[4*nm+3]=x24; yn[4*nm+3]=y24; zn[4*nm+3]=z24; efln[4*nm+3]=efl24;
		    nm += 1;  /* fifth subelement */

		    xn[4*nm+0]=x14; yn[4*nm+0]=y14; zn[4*nm+0]=z14; efln[4*nm+0]=efl14;
		    xn[4*nm+1]=x12; yn[4*nm+1]=y12; zn[4*nm+1]=z12; efln[4*nm+1]=efl12;
		    xn[4*nm+2]=x13; yn[4*nm+2]=y13; zn[4*nm+2]=z13; efln[4*nm+2]=efl13;
		    xn[4*nm+3]=x24; yn[4*nm+3]=y24; zn[4*nm+3]=z24; efln[4*nm+3]=efl24;
		    nm += 1;  /* sixth subelement */

		    xn[4*nm+0]=x12; yn[4*nm+0]=y12; zn[4*nm+0]=z12; efln[4*nm+0]=efl12;
		    xn[4*nm+1]=x23; yn[4*nm+1]=y23; zn[4*nm+1]=z23; efln[4*nm+1]=efl23;
		    xn[4*nm+2]=x13; yn[4*nm+2]=y13; zn[4*nm+2]=z13; efln[4*nm+2]=efl13;
		    xn[4*nm+3]=x24; yn[4*nm+3]=y24; zn[4*nm+3]=z24; efln[4*nm+3]=efl24;
		    nm += 1;  /* seventh subelement */

		    xn[4*nm+0]=x14; yn[4*nm+0]=y14; zn[4*nm+0]=z14; efln[4*nm+0]=efl14;
		    xn[4*nm+1]=x24; yn[4*nm+1]=y24; zn[4*nm+1]=z24; efln[4*nm+1]=efl24;
		    xn[4*nm+2]=x23; yn[4*nm+2]=y23; zn[4*nm+2]=z23; efln[4*nm+2]=efl23;
		    xn[4*nm+3]=x34; yn[4*nm+3]=y34; zn[4*nm+3]=z34; efln[4*nm+3]=efl34;
		    nm += 1;  /* eighth subelement */

		} /* end of loop over current elements */


		ne = 8*ne;  /* number of elements has increased
	                       by a factor of eight */

		for( k = 0; k < ne; ++k )     /* relabel the new points and put them in the master list */
		{                             /* project boundary nodes onto the unit sphere */
		    for( i2 = 0; i2 < 4; ++i2 )
		    {
			x[4*k+i2] = xn[4*k+i2];
			y[4*k+i2] = yn[4*k+i2];
			z[4*k+i2] = zn[4*k+i2];
			efl[4*k+i2] = efln[4*k+i2];
			if( shape_flag == 2 || circle_flag == 1)
			{
			    if(efl[4*k+i2] == 1)
			    {
				rad = sqrt(x[4*k+i2]*x[4*k+i2]+y[4*k+i2]*y[4*k+i2]+z[4*k+i2]*z[4*k+i2]);
				x[4*k+i2] = x[4*k+i2]/rad;
				y[4*k+i2] = y[4*k+i2]/rad;
				z[4*k+i2] = z[4*k+i2]/rad;
			    }
			}
		    }
		}
	    } /* end do of refinement loop */
	} /* end if of refinement loop */

	k = 55;
	i2 = 1;

	/*for( k = 0; k < ne; ++k )
	{
	    for( i2 = 0; i2 < 4; ++i2 )
	    {
		    printf( "mmm %4d %4d %4d %4d %16.8e %16.8e %16.8e %4d\n",
			    3, 128, k+1, i2+1, x[4*k+i2], y[4*k+i2], z[4*k+i2], efl[4*k+i2]);
	    }
	}*/

/*
  define the global nodes and the connectivity table

  three nodes of the first element are entered mannualy
*/

	p[0]=x[0]; p[1]=y[0];  p[2]=z[0];  gfl[0]=efl[0];
	p[3]=x[1]; p[4]=y[1];  p[5]=z[1];  gfl[2]=efl[1];
	p[6]=x[2]; p[7]=y[2];  p[8]=z[2];  gfl[4]=efl[2];
	p[9]=x[3]; p[10]=y[3]; p[11]=z[3]; gfl[6]=efl[3];

	c[0] = 0;  /* first  node of first element is global node 1 */
	c[1] = 1;  /* second node of first element is global node 2 */
	c[2] = 2;  /* third  node of first element is global node 3 */
	c[3] = 3;  /* fourth node of first element is global node 4 */

/* loop over further elements
   Iflag=0 will signal a new global node
*/

	for( i = 1; i < ne; ++i ) /* loop over elements */
	{
	    for( j = 0; j < 4; ++j ) /* loop over element nodes */
	    {
		Iflag=0;
		for( k = 0; k < ng; ++k ) /* loop over element nodes */
		{
		    if( fabs( x[4*i+j] - p[3*k+0] ) < SMALL)
		    {
			if( fabs( y[4*i+j] - p[3*k+1] ) < SMALL)
			{
			    if( fabs( z[4*i+j] - p[3*k+2] ) < SMALL)
			    {
				Iflag = 1;    /* the node has been recorded previously */
				c[npel*i + j] = k;   /* the jth local node of element i is the kth global node */
			    }
			}
		    }
		}

		if( Iflag==0 )  /* record the node */
		{
		    p[3*ng + 0] = x[4*i + j];
		    p[3*ng + 1] = y[4*i + j];
		    p[3*ng + 2] = z[4*i + j];
		    gfl[2*ng + 0] = efl[4*i + j];
		    c[npel*i + j] = ng;   /* the jth local node of element is the new global node */
		    /*printf( "rrr %4d %4d %4d %16.8e %16.8e %16.8e %16.8e %4d\n",
			i, j, ng, x[6*i+j], p[2*k], y[6*i+j], p[2*k+1], c[npel*i+j]);*/
		    ng += 1;
		}
	    }
	} /* end of loop over elements */

	return 1;
}


/*
 Discretize a sphere into 4-node tetrahedral elements
 by successive subdivisions of an 8-element
 parental structure
 
 LEGEND:

 efl: element flag
 gfl: global node flag

 parental structure with eight elements

*/

int main(int argc, char** argv)
{

	int i, j, k, check;
	int dum, dum2;
	double *x, *y, *z, *p, rad;
	int *efl, *c, ndiv, NQ;
	char name[20];
	char buf[ BUFSIZ ];
	FILE *o1;

        shape_flag = 2;
	printf( "Pick the mesh you want among 2\n");
	printf( "1) Cube made up of 768 tetrahedron elements\n");
	printf( "2) Sphere made up of 512 tetrahedron elements\n");
	printf( "3) Use input file\n");
        scanf( "%d", &shape_flag);

	if(shape_flag == 3)
	{
	    ng = 4;
	    printf( "What is the name of the file containing the \n");
	    printf( "initial triangle data? (example: dat1a-te)\n");
	    scanf( "%30s",name);

	    o1 = fopen( name,"r" );
	    if(o1 == NULL ) {
		printf("Can't find file %30s\n",name);
		exit(1);
	    }
	    printf( "   ne ndiv circle_flag\n ");
	    fgets( buf, BUFSIZ, o1 );
	    fscanf( o1, "%d %d %d\n",&ne,&ndiv,&circle_flag);
	    printf( "%4d %4d %4d\n", ne, ndiv, circle_flag);
	}

	if(shape_flag == 2)
	{
/* In  laplt4_d.m, for ndiv = 2 ne = 8 and ng = 4, numel = 512, numnp = 137

2) Second problem:

 CODE laplt4_dXsan.m

 Discretize a sphere into 4-node tetrahedral elements
 by successive subdivisions of an 8-element
 parental structure

 LEGEND:

 efl: element flag
 gfl: global node flag

 parental structure with eight elements

 parental structure with 8 elements
*/
		ne = 8;
		ng = 4;
		ndiv = 2;
		circle_flag = 1;
	}

	if(shape_flag == 1)
	{

/* In tetra4_cube_dr.m, for ndiv = 2 ne = 12 and ng = 4, numel = 768, numnp = 189

1) First problem:

 discretize a cube into 4-node tetrahedral elements
 by successive subdivisions of 12-element
 parental structure

 LEGEND:

 efl: element flag
 gfl: global node flag

 parental structure with 12 elements
*/
		ne=12;
		ng = 4;
		ndiv=2;
	}

	numel = ne*pow(8,ndiv);
	numnp = npel*numel; /* This is temporary so I can set memory.*/
	printf( " %4d %4d\n", numel, numnp);

	x = (double *)calloc(numel*4,sizeof(double));
	y = (double *)calloc(numel*4,sizeof(double));
	z = (double *)calloc(numel*4,sizeof(double));
	efl = (int *)calloc(numel*4,sizeof(int));
	c = (int *)calloc(numel*npel,sizeof(int));
	p = (double *)calloc(numnp*nsd,sizeof(double));

	if(shape_flag == 3)
	{
	    check = tetrareader( efl, name, ndiv, o1, x, y, z);
	    if(!check) printf( " Problems with trglreader\n");
	}

	if(shape_flag == 2)
	{
	    x[0]= 0.0; y[0]= 0.0; z[0]= 0.0; efl[0]=0;  /* first element */
	    x[1]= 1.0; y[1]= 0.0; z[1]= 0.0; efl[1]=1;
	    x[2]= 0.0; y[2]= 1.0; z[2]= 0.0; efl[2]=1;
	    x[3]= 0.0; y[3]= 0.0; z[3]= 1.0; efl[3]=1;

	    x[4]= 0.0; y[4]= 0.0; z[4]= 0.0; efl[4]=0;  /* second element */
	    x[5]= 0.0; y[5]= 1.0; z[5]= 0.0; efl[5]=1;
	    x[6]=-1.0; y[6]= 0.0; z[6]= 0.0; efl[6]=1;
	    x[7]= 0.0; y[7]= 0.0; z[7]= 1.0; efl[7]=1;

	    x[8]= 0.0; y[8]= 0.0; z[8]= 0.0; efl[8]=0;  /* third element */
	    x[9]=-1.0; y[9]= 0.0; z[9]= 0.0; efl[9]=1;
	    x[10]= 0.0; y[10]=-1.0; z[10]= 0.0; efl[10]=1;
	    x[11]= 0.0; y[11]= 0.0; z[11]= 1.0; efl[11]=1;

	    x[12]= 0.0; y[12]= 0.0; z[12]= 0.0; efl[12]=0;  /* fourth element */
	    x[13]= 0.0; y[13]=-1.0; z[13]= 0.0; efl[13]=1;
	    x[14]= 1.0; y[14]= 0.0; z[14]= 0.0; efl[14]=1;
	    x[15]= 0.0; y[15]= 0.0; z[15]= 1.0; efl[15]=1;

	    x[16]= 0.0; y[16]= 0.0; z[16]= 0.0; efl[16]=0;  /* fifth element */
	    x[17]= 0.0; y[17]= 1.0; z[17]= 0.0; efl[17]=1;
	    x[18]= 1.0; y[18]= 0.0; z[18]= 0.0; efl[18]=1;
	    x[19]= 0.0; y[19]= 0.0; z[19]=-1.0; efl[19]=1;

	    x[20]= 0.0; y[20]= 0.0; z[20]= 0.0; efl[20]=0;  /* sixth element */
	    x[21]=-1.0; y[21]= 0.0; z[21]= 0.0; efl[21]=1;
	    x[22]= 0.0; y[22]= 1.0; z[22]= 0.0; efl[22]=1;
	    x[23]= 0.0; y[23]= 0.0; z[23]=-1.0; efl[23]=1;

	    x[24]= 0.0; y[24]= 0.0; z[24]= 0.0; efl[24]=0;  /* seventh element */
	    x[25]= 0.0; y[25]=-1.0; z[25]= 0.0; efl[25]=1;
	    x[26]=-1.0; y[26]= 0.0; z[26]= 0.0; efl[26]=1;
	    x[27]= 0.0; y[27]= 0.0; z[27]=-1.0; efl[27]=1;

	    x[28]= 0.0; y[28]= 0.0; z[28]= 0.0; efl[28]=0;  /* eighth element */
	    x[29]= 1.0; y[29]= 0.0; z[29]= 0.0; efl[29]=1;
	    x[30]= 0.0; y[30]=-1.0; z[30]= 0.0; efl[30]=1;
	    x[31]= 0.0; y[31]= 0.0; z[31]=-1.0; efl[31]=1;
	}


	if(shape_flag == 1)
	{
	    x[0]= 0.0; y[0]= 0.0; z[0]= 0.0; efl[0]=0;  /* first element */
	    x[1]= 1.0; y[1]=-1.0; z[1]= 1.0; efl[1]=1;
	    x[2]= 1.0; y[2]= 1.0; z[2]=-1.0; efl[2]=1;
	    x[3]= 1.0; y[3]= 1.0; z[3]= 1.0; efl[3]=1;

	    x[4]= 0.0; y[4]= 0.0; z[4]= 0.0; efl[4]=0;  /* second element */
	    x[5]= 1.0; y[5]=-1.0; z[5]= 1.0; efl[5]=1;
	    x[6]= 1.0; y[6]=-1.0; z[6]=-1.0; efl[6]=1;
	    x[7]= 1.0; y[7]= 1.0; z[7]=-1.0; efl[7]=1;

	    x[8]= 0.0; y[8]= 0.0; z[8]= 0.0; efl[8]=0;  /* third element */
	    x[9]= 1.0; y[9]=-1.0; z[9]= 1.0; efl[9]=1;
	    x[10]=-1.0; y[10]=-1.0; z[10]= 1.0; efl[10]=1;
	    x[11]= 1.0; y[11]=-1.0; z[11]=-1.0; efl[11]=1;

	    x[12]= 0.0; y[12]= 0.0; z[12]= 0.0; efl[12]=0;  /* fourth element */
	    x[13]=-1.0; y[13]=-1.0; z[13]= 1.0; efl[13]=1;
	    x[14]=-1.0; y[14]=-1.0; z[14]=-1.0; efl[14]=1;
	    x[15]= 1.0; y[15]=-1.0; z[15]=-1.0; efl[15]=1;

	    x[16]= 0.0; y[16]= 0.0; z[16]= 0.0; efl[16]=0;  /* fifth element */
	    x[17]=-1.0; y[17]= 1.0; z[17]= 1.0; efl[17]=1;
	    x[18]= 1.0; y[18]= 1.0; z[18]= 1.0; efl[18]=1;
	    x[19]= 1.0; y[19]= 1.0; z[19]=-1.0; efl[19]=1;

	    x[20]= 0.0; y[20]= 0.0; z[20]= 0.0; efl[20]=0;  /* sixth element */
	    x[21]=-1.0; y[21]= 1.0; z[21]=-1.0; efl[21]=1;
	    x[22]=-1.0; y[22]= 1.0; z[22]= 1.0; efl[22]=1;
	    x[23]= 1.0; y[23]= 1.0; z[23]=-1.0; efl[23]=1;

	    x[24]= 0.0; y[24]= 0.0; z[24]= 0.0; efl[24]=0;  /* seventh element */
	    x[25]=-1.0; y[25]= 1.0; z[25]=-1.0; efl[25]=1;
	    x[26]=-1.0; y[26]=-1.0; z[26]= 1.0; efl[26]=1;
	    x[27]=-1.0; y[27]= 1.0; z[27]= 1.0; efl[27]=1;

	    x[28]= 0.0; y[28]= 0.0; z[28]= 0.0; efl[28]=0;  /* eighth element */
	    x[29]=-1.0; y[29]=-1.0; z[29]=-1.0; efl[29]=1;
	    x[30]=-1.0; y[30]=-1.0; z[30]= 1.0; efl[30]=1;
	    x[31]=-1.0; y[31]= 1.0; z[31]=-1.0; efl[31]=1;

	    x[32]= 0.0; y[32]= 0.0; z[32]= 0.0; efl[32]=0;  /* ninth element */
	    x[33]=-1.0; y[33]=-1.0; z[33]= 1.0; efl[33]=1;
	    x[34]= 1.0; y[34]=-1.0; z[34]= 1.0; efl[34]=1;
	    x[35]=-1.0; y[35]= 1.0; z[35]= 1.0; efl[35]=1;

	    x[36]= 0.0; y[36]= 0.0; z[36]= 0.0; efl[36]=0;  /* tenth element */
	    x[37]=-1.0; y[37]= 1.0; z[37]= 1.0; efl[37]=1;
	    x[38]= 1.0; y[38]=-1.0; z[38]= 1.0; efl[38]=1;
	    x[39]= 1.0; y[39]= 1.0; z[39]= 1.0; efl[39]=1;

	    x[40]= 0.0; y[40]= 0.0; z[40]= 0.0; efl[40]=0;  /* eleventh element */
	    x[41]= 1.0; y[41]=-1.0; z[41]=-1.0; efl[41]=1;
	    x[42]=-1.0; y[42]=-1.0; z[42]=-1.0; efl[42]=1;
	    x[43]=-1.0; y[43]= 1.0; z[43]=-1.0; efl[43]=1;

	    x[44]= 0.0; y[44]= 0.0; z[44]= 0.0; efl[44]=0;  /* twelfth element */
	    x[45]= 1.0; y[45]=-1.0; z[45]=-1.0; efl[45]=1;
	    x[46]=-1.0; y[46]= 1.0; z[46]=-1.0; efl[46]=1;
	    x[47]= 1.0; y[47]= 1.0; z[47]=-1.0; efl[47]=1;
	}

	dum = 0;
	printf( "numel    ndiv   circle flag  radius\n");
	printf( "%4d %4d %4d\n", ne, ndiv, circle_flag);
	printf( "element   x   y   z   efl\n");
	for( i = 0; i < ne; ++i )
	{
	    printf( "%d\n", i);
	    for( j = 0; j < npel; ++j )
	    {
		printf( "   %16.8e %16.8e %16.8e %4d", x[dum], y[dum], z[dum], efl[dum]);
		printf( "\n");
		++dum;
	    }
	}

	check = tetra4( x, y, z, p, efl, c, ndiv);
	if(!check) printf( " Problems with trgl3_sqr \n");

	numnp = ng;
	numel = ne;

	check = slffeawrite( p, c, npel, nsd);
	if(!check) printf( " Problems with slffeawrite \n");

	return 1;

}


