/*
    This program generates meshes for triangle elements.  It does it by
    sub-division.  It is taken from the Matlab functions trgl6_rc.m,
    trgl6_sc.m, trgl6_ss.m, trgl6_sqr.m, trgl6_disk.m in the package FSELIB
    for chapters 4 and 5 written by C. Pozrikidis from his book:

 Pozrikidis, C., Introduction to Finite and Spectral Element Methods using Matlab,
      Chapman & Hall/CRC, Boca Raton, 2005.

    You can get the complete FSELIB package at:

      http://dehesa.freeshell.org/FSELIB/

    Note that although this code generates 6 node triangles, I will only write out
    the first 3 nodes which define the verticies of the triangle and not
    the ones on the midsides. 

                Updated 1/9/07 Translated into C by San Le

                Copyright C. Pozrikidis

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define SMALL   0.000001


int ne, ng, ndiv;
int numnp, numel, shape_flag, triangle_flag = 1, circle_flag;
int npel = 6, nsd = 2;

int trglreader(int *, char *, int , FILE *, double *, double *);

int slffeawrite(double *, int *, int, int);

int trgl6( double *x, double *y, double *p, int *efl, int *c, int ndiv, double a)
{
	int i,i2,j,k,dum,dum2,dum3,dum4;
	double fdum, fdum1, fdum2, fdum3, fdum4, Iflag;
	double *xn, *yn, xp[40], yp[40], rad;
	int *efln, *gfl;

	int nm;

	xn = (double *)calloc(numel*4*ndiv*6,sizeof(double));
	yn = (double *)calloc(numel*4*ndiv*6,sizeof(double));
	efln = (int *)calloc(numel*4*ndiv*6,sizeof(int));
	gfl = (int *)calloc(numnp*3,sizeof(int));

/*
 refinement loop
*/

	if(ndiv > 0)
	{
	    for( i = 0; i < ndiv; ++i )
	    {
		nm = 0; /* count the new elements arising by each refinement loop
			   four elements will be generated in each pass */

		for( j = 0; j < ne; ++j )   /* loop over current elements */
		{

/* assign vertex nodes to sub-elements
   these will become the "new" elements
*/

		    xn[6*nm+0]=x[6*j+0]; yn[6*nm+0]=y[6*j+0]; efln[6*nm+0]=efl[6*j+0]; /*  first sub-element */
		    xn[6*nm+1]=x[6*j+3]; yn[6*nm+1]=y[6*j+3]; efln[6*nm+1]=efl[6*j+3];
		    xn[6*nm+2]=x[6*j+5]; yn[6*nm+2]=y[6*j+5]; efln[6*nm+2]=efl[6*j+5];
		    xn[6*nm+3] = 0.5*(xn[6*nm+0]+xn[6*nm+1]);yn[6*nm+3] = 0.5*(yn[6*nm+0]+yn[6*nm+1]);
		    xn[6*nm+4] = 0.5*(xn[6*nm+1]+xn[6*nm+2]);yn[6*nm+4] = 0.5*(yn[6*nm+1]+yn[6*nm+2]);
		    xn[6*nm+5] = 0.5*(xn[6*nm+2]+xn[6*nm+0]);yn[6*nm+5] = 0.5*(yn[6*nm+2]+yn[6*nm+0]);

		    efln[6*nm+3] = 0; if(efln[6*nm+0]==1 & efln[6*nm+1]==1) efln[6*nm+3] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+0]==2 & efln[6*nm+1]==2) efln[6*nm+3] = 2; }
		    efln[6*nm+4] = 0; if(efln[6*nm+1]==1 & efln[6*nm+2]==1) efln[6*nm+4] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+1]==2 & efln[6*nm+2]==2) efln[6*nm+4] = 2; }
		    efln[6*nm+5] = 0; if(efln[6*nm+2]==1 & efln[6*nm+0]==1) efln[6*nm+5] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+2]==2 & efln[6*nm+0]==2) efln[6*nm+5] = 2; }
		    nm = nm+1;


		    xn[6*nm+0]=x[6*j+3]; yn[6*nm+0]=y[6*j+3]; efln[6*nm+0]=efl[6*j+3]; /*  second sub-element */
		    xn[6*nm+1]=x[6*j+1]; yn[6*nm+1]=y[6*j+1]; efln[6*nm+1]=efl[6*j+1];
		    xn[6*nm+2]=x[6*j+4]; yn[6*nm+2]=y[6*j+4]; efln[6*nm+2]=efl[6*j+4];
		    xn[6*nm+3] = 0.5*(xn[6*nm+0]+xn[6*nm+1]);yn[6*nm+3] = 0.5*(yn[6*nm+0]+yn[6*nm+1]);
		    xn[6*nm+4] = 0.5*(xn[6*nm+1]+xn[6*nm+2]);yn[6*nm+4] = 0.5*(yn[6*nm+1]+yn[6*nm+2]);
		    xn[6*nm+5] = 0.5*(xn[6*nm+2]+xn[6*nm+0]);yn[6*nm+5] = 0.5*(yn[6*nm+2]+yn[6*nm+0]);

		    efln[6*nm+3] = 0; if(efln[6*nm+0]==1 & efln[6*nm+1]==1) efln[6*nm+3] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+0]==2 & efln[6*nm+1]==2) efln[6*nm+3] = 2; }
		    efln[6*nm+4] = 0; if(efln[6*nm+1]==1 & efln[6*nm+2]==1) efln[6*nm+4] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+1]==2 & efln[6*nm+2]==2) efln[6*nm+4] = 2; }
		    efln[6*nm+5] = 0; if(efln[6*nm+2]==1 & efln[6*nm+0]==1) efln[6*nm+5] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+2]==2 & efln[6*nm+0]==2) efln[6*nm+5] = 2; }
		    nm = nm+1;


		    xn[6*nm+0]=x[6*j+5]; yn[6*nm+0]=y[6*j+5]; efln[6*nm+0]=efl[6*j+5]; /*  third sub-element */
		    xn[6*nm+1]=x[6*j+4]; yn[6*nm+1]=y[6*j+4]; efln[6*nm+1]=efl[6*j+4];
		    xn[6*nm+2]=x[6*j+2]; yn[6*nm+2]=y[6*j+2]; efln[6*nm+2]=efl[6*j+2];
		    xn[6*nm+3] = 0.5*(xn[6*nm+0]+xn[6*nm+1]);yn[6*nm+3] = 0.5*(yn[6*nm+0]+yn[6*nm+1]);
		    xn[6*nm+4] = 0.5*(xn[6*nm+1]+xn[6*nm+2]);yn[6*nm+4] = 0.5*(yn[6*nm+1]+yn[6*nm+2]);
		    xn[6*nm+5] = 0.5*(xn[6*nm+2]+xn[6*nm+0]);yn[6*nm+5] = 0.5*(yn[6*nm+2]+yn[6*nm+0]);

		    efln[6*nm+3] = 0; if(efln[6*nm+0]==1 & efln[6*nm+1]==1) efln[6*nm+3] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+0]==2 & efln[6*nm+1]==2) efln[6*nm+3] = 2; }
		    efln[6*nm+4] = 0; if(efln[6*nm+1]==1 & efln[6*nm+2]==1) efln[6*nm+4] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+1]==2 & efln[6*nm+2]==2) efln[6*nm+4] = 2; }
		    efln[6*nm+5] = 0; if(efln[6*nm+2]==1 & efln[6*nm+0]==1) efln[6*nm+5] = 1;
		    if( shape_flag < 4 || circle_flag == 2) { if(efln[6*nm+2]==2 & efln[6*nm+0]==2) efln[6*nm+5] = 2; }
		    nm = nm+1;


		    xn[6*nm+0]=x[6*j+3]; yn[6*nm+0]=y[6*j+3]; efln[6*nm+0]=efl[6*j+3]; /*  fourth sub-element */
		    xn[6*nm+1]=x[6*j+4]; yn[6*nm+1]=y[6*j+4]; efln[6*nm+1]=efl[6*j+4];
		    xn[6*nm+2]=x[6*j+5]; yn[6*nm+2]=y[6*j+5]; efln[6*nm+2]=efl[6*j+5];
		    xn[6*nm+3] = 0.5*(xn[6*nm+0]+xn[6*nm+1]);yn[6*nm+3] = 0.5*(yn[6*nm+0]+yn[6*nm+1]);
		    xn[6*nm+4] = 0.5*(xn[6*nm+1]+xn[6*nm+2]);yn[6*nm+4] = 0.5*(yn[6*nm+1]+yn[6*nm+2]);
		    xn[6*nm+5] = 0.5*(xn[6*nm+2]+xn[6*nm+0]);yn[6*nm+5] = 0.5*(yn[6*nm+2]+yn[6*nm+0]);

		    if( shape_flag < 4 || circle_flag == 2)
		    {
			efln[6*nm+3] = 0; efln[6*nm+4] = 0; efln[6*nm+5] = 0; 
			if(efln[6*nm+0]==1 & efln[6*nm+1]==1) efln[6*nm+3] = 1;
			if(efln[6*nm+0]==2 & efln[6*nm+1]==2) efln[6*nm+3] = 2;
			if(efln[6*nm+1]==1 & efln[6*nm+2]==1) efln[6*nm+4] = 1;
			if(efln[6*nm+1]==2 & efln[6*nm+2]==2) efln[6*nm+4] = 2;
			if(efln[6*nm+2]==1 & efln[6*nm+0]==1) efln[6*nm+5] = 1;
			if(efln[6*nm+2]==2 & efln[6*nm+0]==2) efln[6*nm+5] = 2;
		    }
		    if( shape_flag > 3 && shape_flag < 6 || circle_flag == 1 )
		    {
			efln[6*nm+3] = 0; if(efln[6*nm+0]==1 & efln[6*nm+1]==1) efln[6*nm+3] = 1;
			efln[6*nm+4] = 0; if(efln[6*nm+1]==1 & efln[6*nm+2]==1) efln[6*nm+4] = 1;
			efln[6*nm+5] = 0; if(efln[6*nm+2]==1 & efln[6*nm+0]==1) efln[6*nm+5] = 1;
		    }
		    nm = nm+1;

		} /* end of loop over current elements */


		ne = 4*ne;  /* number of elements has increased
			       by a factor of four */

		for( k = 0; k < ne; ++k ) /* relabel the new points and put them in the master list */
		{
		    for( i2 = 0; i2 < 6; ++i2 ) /* relabel the new points and put them in the master list */
		    {
			x[6*k+i2] = xn[6*k+i2];
			y[6*k+i2] = yn[6*k+i2];
			efl[6*k+i2] = efln[6*k+i2];
			if ( shape_flag > 1 && shape_flag < 4 || circle_flag == 2 )
			{
			    if( efl[6*k+i2] == 2 )
			    {
				rad = a/sqrt(x[6*k+i2]*x[6*k+i2] + y[6*k+i2]*y[6*k+i2]);
				x[6*k+i2] = x[6*k+i2]*rad;
				y[6*k+i2] = y[6*k+i2]*rad;
			    }
			}
			if ( shape_flag == 4 || circle_flag == 1)
			{
			    if( efl[6*k+i2] == 1 )
			    {
				rad = sqrt(x[6*k+i2]*x[6*k+i2] + y[6*k+i2]*y[6*k+i2]);
				x[6*k+i2] = x[6*k+i2]/rad;
				y[6*k+i2] = y[6*k+i2]/rad;
			    }

			}
		    }
		}
		if ( shape_flag == 4 || circle_flag == 1)
		{
		    for( k = 0; k < ne; ++k ) /* reposition the mid-nodes */
		    {
			if(efl[6*k+3] == 0)
			    x[6*k+3]=0.5*(x[6*k+0]+x[6*k+1]);
			    y[6*k+3]=0.5*(y[6*k+0]+y[6*k+1]);
			if(efl[6*k+4] == 0)
			    x[6*k+4]=0.5*(x[6*k+1]+x[6*k+2]);
			    y[6*k+4]=0.5*(y[6*k+1]+y[6*k+2]);
			if(efl[6*k+5] == 0)
			    x[6*k+5]=0.5*(x[6*k+2]+x[6*k+0]);
			    y[6*k+5]=0.5*(y[6*k+2]+y[6*k+0]);
		    }
		}
	    } /* end do of refinement loop */
	} /* end if of refinement loop */

/*
 define the global nodes and the connectivity table
*/
	p[0]=x[0];  p[1]=y[0];  gfl[0]=efl[0];
	p[2]=x[1];  p[3]=y[1];  gfl[2]=efl[1];
	p[4]=x[2];  p[5]=y[2];  gfl[4]=efl[2];
	p[6]=x[3];  p[7]=y[3];  gfl[6]=efl[3];
	p[8]=x[4];  p[9]=y[4];  gfl[8]=efl[4];
	p[10]=x[5]; p[11]=y[5]; gfl[10]=efl[5];

	c[0] = 0;  /* first  node of first element is global node 1 */
	c[1] = 1;  /* second node of first element is global node 2 */
	c[2] = 2;  /* third  node of first element is global node 3 */
	c[3] = 3;  /* fourth node of first element is global node 4 */
	c[4] = 4;  /* fifth  node of first element is global node 5 */
	c[5] = 5;  /* sixth  node of first element is global node 6 */

/* loop over further elements
   Iflag=0 will signal a new global node
*/

	for( i = 1; i < ne; ++i ) /* loop over elements */
	{
	    for( j = 0; j < 6; ++j ) /* loop over element nodes */
	    {
		Iflag=0;
		for( k = 0; k < ng; ++k ) /* loop over element nodes */
		{
		    if( fabs( x[6*i+j] - p[2*k+0] ) < SMALL)
		    {
			if( fabs( y[6*i+j] - p[2*k+1] ) < SMALL)
			{

			    Iflag = 1;    /* the node has been recorded previously */
			    c[npel*i + j] = k;   /* the jth local node of element i is the kth global node */
			}
		    }
		}

		if( Iflag==0 )  /* record the node */
		{
		    p[2*ng + 0] = x[6*i + j];
		    p[2*ng + 1] = y[6*i + j];
		    gfl[2*ng + 0] = efl[6*i + j];
		    c[npel*i + j] = ng;   /* the jth local node of element is the new global node */
                    /*printf( "rrr %4d %4d %4d %16.8e %16.8e %16.8e %16.8e %4d\n",
                        i, j, ng, x[6*i+j], p[2*k], y[6*i+j], p[2*k+1], c[npel*i+j]);*/
		    ng += 1;
		}
	    }
	} /* end of loop over elements */

	return 1;
}

/*
 FSELIB
*/

int main(int argc, char** argv)
{

	int i, j, k, check; 
	int dum, dum2;
	double *x, *y, *p, a, rad;
	int *efl, *c, ndiv, NQ;
	char name[20];
	char buf[ BUFSIZ ];
	FILE *o1;

	a = 0.0;

	shape_flag = 5;
	printf( "Pick the mesh you want among 5\n");
	printf( "1) Square with square hole made up of 48 triangle elements\n");
	printf( "2) Square with circular hole made up of 192 triangle elements\n");
	printf( "3) 2 squares, one with circular hole made up of 320 triangle elements\n");
	printf( "4) Disk made up of 256 triangle elements\n");
	printf( "5) Square made up of 128 triangle elements\n");
	printf( "6) Use input file\n");
	scanf( "%d", &shape_flag);

	if(shape_flag == 6)
	{
	    ng = 6;
	    printf( "What is the name of the file containing the \n");
	    printf( "initial triangle data? (example: dat3a-6)\n");
	    scanf( "%30s",name);

	    o1 = fopen( name,"r" );
	    if(o1 == NULL ) {
		printf("Can't find file %30s\n",name);
		exit(1);
	    }
	    printf( "   ne ndiv circle_flag\n ");
	    fgets( buf, BUFSIZ, o1 );
	    fscanf( o1, "%d %d %d %lf\n",&ne,&ndiv,&circle_flag, &a);
	    printf( "%4d %4d %4d %16.8e\n", ne, ndiv, circle_flag, a);
	}

	if(shape_flag == 5)
	{
/* In membrane.m, for ndiv = 2 ne = 8 and ng = 6, numel = 128, numnp = 289

   Triangulation of the unit disk into 6-node elements
   by the successive subdivision
   of four parental elements.

   parental structure with 8 elements
*/
		ne = 8;
		ng = 6;
		ndiv = 2;
		NQ = 6;    /* gauss-triangle quadrature */
		circle_flag = 0;

	}

	if(shape_flag == 4)
	{
/* In pois6_d.m, for ndiv = 3 ne = 4 and ng = 6, numel = 256, numnp = 545

   Triangulation of the unit disk into 6-node elements
   by the successive subdivision
   of four parental elements.

   parental structure with 4 elements
*/
		ne = 4;
		ng = 6;
		ndiv = 3;
		NQ = 6;    /* gauss-triangle quadrature */
		circle_flag = 1;
	}

	if(shape_flag == 3)
	{
/* In lapl6_d_rc.m, for ndiv = 2 ne = 20 and ng = 6, numel = 320, numnp = 704

3) Third problem:

  Triangulation of a rectangle with a circular hole
  of radius ''a'' into 6-node elements
  by the successive subdivision of 20 parental elements.

*/
		ne = 20;
		ng = 6;
		ndiv = 2;
		NQ = 6;    /* gauss-triangle quadrature */
		a = 0.5;
		circle_flag = 2;
	}

	if(shape_flag == 2)
	{
/* In  lapl6_d_sc.m, for ndiv = 2 ne = 12 and ng = 6, numel = 192, numnp = 432

2) Second problem:

 Triangulation of a unit square with a circular hole
 of radius ``a'' into 6-node elements
 by the successive subdivision of 12 parental elements.

 parental structure with 12 elements

*/
		ne = 12;
		ng = 6;
		ndiv = 2;
		NQ = 6;    /* gauss-triangle quadrature */
		a = 0.5;
		circle_flag = 2;
	}

	if(shape_flag == 1)
	{
/* In lapl6_d_ss.m, for ndiv = 1 ne = 12 and ng = 6, numel = 48, numnp = 120

1) First problem:

 Triangulation of a unit square with a square hole
 of semi-side ''a'' into 6-node elements
 by the successive subdivision of 12 parental elements.

 parental structure with 12 elements
*/
		ne = 12;
		ng = 6;
		ndiv = 1;
		NQ = 6;    /* gauss-triangle quadrature */
		a = 0.4;
		circle_flag = 0;
	}

	numel = ne*pow(4,ndiv);
	numnp = npel*numel; /* This is temporary so I can set memory.*/
	printf( " %4d %4d\n", numel, numnp);

	x = (double *)calloc(numel*6,sizeof(double));
	y = (double *)calloc(numel*6,sizeof(double));
	efl = (int *)calloc(numel*npel,sizeof(int));
	c = (int *)calloc(numel*6,sizeof(int));
	p = (double *)calloc(numnp*nsd,sizeof(double));

/* triangulate */

	if(shape_flag == 6)
	{
	    check = trglreader( efl, name, ndiv, o1, x, y);
	    if(!check) printf( " Problems with trglreader\n");
	}

	if( shape_flag < 4)
	{
	    x[0] =   a; y[0] =  -a; efl[0] = 2;  /* first element */
	    x[1] = 1.0; y[1] =-1.0; efl[1] = 1;
	    x[2] = 1.0; y[2] = 0.0; efl[2] = 0;
	    x[3] = 0.5*(x[0]+x[1]); y[3] = 0.5*(y[0]+y[1]); efl[3] = 0;
	    x[4] = 0.5*(x[1]+x[2]); y[4] = 0.5*(y[1]+y[2]); efl[4] = 0;
	    x[5] = 0.5*(x[2]+x[0]); y[5] = 0.5*(y[2]+y[0]); efl[5] = 0;

	    x[6] =   a; y[6] =  -a; efl[6] = 2;  /* second element */
	    x[7] = 1.0; y[7] = 0.0; efl[7] = 0;
	    x[8] =   a; y[8] =   a; efl[8] = 2;
	    x[9] = 0.5*(x[6]+x[7]); y[9] = 0.5*(y[6]+y[7]); efl[9] = 0;
	    x[10] = 0.5*(x[7]+x[8]); y[10] = 0.5*(y[7]+y[8]); efl[10] = 0;
	    x[11] = 0.5*(x[8]+x[6]); y[11] = 0.5*(y[8]+y[6]); efl[11] = 2;

	    x[12] =   a; y[12] =   a; efl[12] = 2;  /* third element */
	    x[13] = 1.0; y[13] = 0.0; efl[13] = 0;
	    x[14] = 1.0; y[14] = 1.0; efl[14] = 1;
	    x[15] = 0.5*(x[12]+x[13]); y[15] = 0.5*(y[12]+y[13]); efl[15] = 0;
	    x[16] = 0.5*(x[13]+x[14]); y[16] = 0.5*(y[13]+y[14]); efl[16] = 0;
	    x[17] = 0.5*(x[14]+x[12]); y[17] = 0.5*(y[14]+y[12]); efl[17] = 0;

	    x[18] =   a; y[18] =   a; efl[18] = 2;  /* fourth element */
	    x[19] = 1.0; y[19] = 1.0; efl[19] = 1;
	    x[20] = 0.0; y[20] = 1.0; efl[20] = 1;
	    x[21] = 0.5*(x[18]+x[19]); y[21] = 0.5*(y[18]+y[19]); efl[21] = 0;
	    x[22] = 0.5*(x[19]+x[20]); y[22] = 0.5*(y[19]+y[20]); efl[22] = 1;
	    x[23] = 0.5*(x[20]+x[18]); y[23] = 0.5*(y[20]+y[18]); efl[23] = 0;

	    x[24] = 0.0; y[24] = 1.0; efl[24] = 1;  /* fifth element */
	    x[25] =  -a; y[25] =   a; efl[25] = 2;
	    x[26] =   a; y[26] =   a; efl[26] = 2;
	    x[27] = 0.5*(x[24]+x[25]); y[27] = 0.5*(y[24]+y[25]); efl[27] = 0;
	    x[28] = 0.5*(x[25]+x[26]); y[28] = 0.5*(y[25]+y[26]); efl[28] = 2;
	    x[29] = 0.5*(x[26]+x[24]); y[29] = 0.5*(y[26]+y[24]); efl[29] = 0;

	    x[30] =  -a; y[30] =   a; efl[30] = 2;  /* sixth element */
	    x[31] = 0.0; y[31] = 1.0; efl[31] = 1;
	    x[32] =-1.0; y[32] = 1.0; efl[32] = 1;
	    x[33] = 0.5*(x[30]+x[31]); y[33] = 0.5*(y[30]+y[31]); efl[33] = 0;
	    x[34] = 0.5*(x[31]+x[32]); y[34] = 0.5*(y[31]+y[32]); efl[34] = 1;
	    x[35] = 0.5*(x[32]+x[30]); y[35] = 0.5*(y[32]+y[30]); efl[35] = 0;

/*
     reflect to generate elements: 7-12
*/

	    for ( i = 0; i < 6; ++i )
	    {
		for ( j = 0; j < 6; ++j )
		{
		    x[6*(6+i)+j] = -x[6*i+j];
		    y[6*(6+i)+j] = -y[6*i+j];
		    efl[6*(6+i)+j] = efl[6*i+j];
		}
	    }

	    efl[36] = 2; efl[37] = 1; efl[38] = 1; efl[39] = 0; efl[40] = 1; efl[41] = 0;
	    efl[42] = 2; efl[43] = 1; efl[44] = 2; efl[45] = 0; efl[46] = 0; efl[47] = 2;
	    efl[48] = 2; efl[49] = 1; efl[50] = 1; efl[51] = 0; efl[52] = 1; efl[53] = 0;
	    efl[54] = 2; efl[55] = 1; efl[56] = 1; efl[57] = 0; efl[58] = 1; efl[59] = 0;
	    efl[60] = 1; efl[61] = 2; efl[62] = 2; efl[63] = 0; efl[64] = 2; efl[65] = 0;
	    efl[66] = 2; efl[67] = 1; efl[68] = 1; efl[69] = 0; efl[70] = 1; efl[71] = 0;

	    if(shape_flag > 1 )
	    {
/* project on the circle */

		for ( i = 0; i < 12; ++i )
		{
		    for ( j = 0; j < 6; ++j )
		    {
			if ( efl[6*i+j] == 2 )
			{
			    rad = a/sqrt( x[6*i+j]*x[6*i+j] + y[6*i+j]*y[6*i+j]);
			    x[6*i+j] = x[6*i+j]*rad;
			    y[6*i+j] = y[6*i+j]*rad;
			}
		    }
		}

		if(shape_flag > 2 )
		{

/*     elements on the right part */

		    x[72] = 1.0; y[72] =  -1.0; efl[72] = 1;  /* element 13 */
		    x[73] = 2.0; y[73] =  -1.0; efl[73] = 1; 
		    x[74] = 1.0; y[74] =   0.0; efl[74] = 0;
		    x[75] = 0.5*(x[72]+x[73]);y[75] = 0.5*(y[72]+y[73]); efl[75] = 1;
		    x[76] = 0.5*(x[73]+x[74]);y[76] = 0.5*(y[73]+y[74]); efl[76] = 0;
		    x[77] = 0.5*(x[74]+x[72]);y[77] = 0.5*(y[74]+y[72]); efl[77] = 0;

		    x[78] = 2.0;y[78] =  -1.0; efl[78] = 1;  /* element 14 */
		    x[79] = 2.0;y[79] =  -0.0; efl[79] = 0;
		    x[80] = 1.0;y[80] =   0.0; efl[80] = 0;
		    x[81] = 0.5*(x[78]+x[79]);y[81] = 0.5*(y[78]+y[79]); efl[81] = 0;
		    x[82] = 0.5*(x[79]+x[80]);y[82] = 0.5*(y[79]+y[80]); efl[82] = 0;
		    x[83] = 0.5*(x[80]+x[78]);y[83] = 0.5*(y[80]+y[78]); efl[83] = 0;

		    x[84] = 2.0;y[84] =   0.0; efl[84] = 0;  /* element 15 */
		    x[85] = 2.0;y[85] =   1.0; efl[85] = 1;
		    x[86] = 1.0;y[86] =   0.0; efl[86] = 0; 
		    x[87] = 0.5*(x[84]+x[85]);y[87] = 0.5*(y[84]+y[85]); efl[87] = 0;
		    x[88] = 0.5*(x[85]+x[86]);y[88] = 0.5*(y[85]+y[86]); efl[88] = 0;
		    x[89] = 0.5*(x[86]+x[84]);y[89] = 0.5*(y[86]+y[84]); efl[89] = 0;

		    x[90] = 1.0;y[90] =   0.0; efl[90] = 0;  /* element 16 */
		    x[91] = 2.0;y[91] =   1.0; efl[91] = 1; 
		    x[92] = 1.0;y[92] =   1.0; efl[92] = 1; 
		    x[93] = 0.5*(x[90]+x[91]);y[93] = 0.5*(y[90]+y[91]); efl[93] = 0;
		    x[94] = 0.5*(x[91]+x[92]);y[94] = 0.5*(y[91]+y[92]); efl[94] = 1;
		    x[95] = 0.5*(x[92]+x[90]);y[95] = 0.5*(y[92]+y[90]); efl[95] = 0;

		    x[96] = 2.0;y[96] =  -1.0; efl[96] = 1;  /* element 17 */
		    x[97] = 3.0;y[97] =  -1.0; efl[97] = 1;
		    x[98] = 2.0;y[98] =   0.0; efl[98] = 0;
		    x[99] = 0.5*(x[96]+x[97]);y[99] = 0.5*(y[96]+y[97]); efl[99] = 1;
		    x[100] = 0.5*(x[97]+x[98]);y[100] = 0.5*(y[97]+y[98]); efl[100] = 0;
		    x[101] = 0.5*(x[98]+x[96]);y[101] = 0.5*(y[98]+y[96]); efl[101] = 0;

		    x[102] = 3.0;y[102] =  -1.0; efl[102] = 1;  /* element 18 */
		    x[103] = 3.0;y[103] =   0.0; efl[103] = 1;
		    x[104] = 2.0;y[104] =   0.0; efl[104] = 0; 
		    x[105] = 0.5*(x[102]+x[103]);y[105] = 0.5*(y[102]+y[103]); efl[105] = 1;
		    x[106] = 0.5*(x[103]+x[104]);y[106] = 0.5*(y[103]+y[104]); efl[106] = 0;
		    x[107] = 0.5*(x[104]+x[102]);y[107] = 0.5*(y[104]+y[102]); efl[107] = 0;

		    x[108] = 3.0;y[108] =   0.0; efl[108] = 1;  /* element 19 */
		    x[109] = 3.0;y[109] =   1.0; efl[109] = 1;
		    x[110] = 2.0;y[110] =   0.0; efl[110] = 0;
		    x[111] = 0.5*(x[108]+x[109]);y[111] = 0.5*(y[108]+y[109]); efl[111] = 1;
		    x[112] = 0.5*(x[109]+x[110]);y[112] = 0.5*(y[109]+y[110]); efl[112] = 0;
		    x[113] = 0.5*(x[110]+x[108]);y[113] = 0.5*(y[110]+y[108]); efl[113] = 0;

		    x[114] = 3.0;y[114] =   1.0; efl[114] = 1;  /* element 20 */
		    x[115] = 2.0;y[115] =   1.0; efl[115] = 1; 
		    x[116] = 2.0;y[116] =   0.0; efl[116] = 0; 
		    x[117] = 0.5*(x[114]+x[115]);y[117] = 0.5*(y[114]+y[115]); efl[117] = 1;
		    x[118] = 0.5*(x[115]+x[116]);y[118] = 0.5*(y[115]+y[116]); efl[118] = 0;
		    x[119] = 0.5*(x[116]+x[114]);y[119] = 0.5*(y[116]+y[114]); efl[119] = 0;
		}
	    }
	}

	if(shape_flag == 4)
	{
	    x[0] = 0.0; y[0] = 0.0; efl[0]=0;  /* first element */
	    x[1] = 1.0; y[1] = 0.0; efl[1]=1;
	    x[2] = 0.0; y[2] = 1.0; efl[2]=1;
	    x[3] = 0.5*(x[0]+x[1]); y[3] = 0.5*(y[0]+y[1]); efl[3]=0;
	    x[4] = 0.5*(x[1]+x[2]); y[4] = 0.5*(y[1]+y[2]); efl[4]=1;
	    x[5] = 0.5*(x[2]+x[0]); y[5] = 0.5*(y[2]+y[0]); efl[5]=0;

	    x[6] = 0.0; y[6] = 0.0; efl[6]=0;  /* second element */
	    x[7] = 0.0; y[7] = 1.0; efl[7]=1;
	    x[8] =-1.0; y[8] = 0.0; efl[8]=1;
	    x[9] = 0.5*(x[6]+x[7]); y[9] = 0.5*(y[6]+y[7]); efl[9]=0;
	    x[10] = 0.5*(x[7]+x[8]); y[10] = 0.5*(y[7]+y[8]); efl[10]=1;
	    x[11] = 0.5*(x[8]+x[6]); y[11] = 0.5*(y[8]+y[6]); efl[11]=0;

	    x[12] = 0.0; y[12] = 0.0; efl[12]=0;  /* third element */
	    x[13] =-1.0; y[13] = 0.0; efl[13]=1;
	    x[14] = 0.0; y[14] =-1.0; efl[14]=1;
	    x[15] = 0.5*(x[12]+x[13]); y[15] = 0.5*(y[12]+y[13]); efl[15]=0;
	    x[16] = 0.5*(x[13]+x[14]); y[16] = 0.5*(y[13]+y[14]); efl[16]=1;
	    x[17] = 0.5*(x[14]+x[12]); y[17] = 0.5*(y[14]+y[12]); efl[17]=0;

	    x[18] = 0.0; y[18] = 0.0; efl[18]=0;  /* fourth element */
	    x[19] = 0.0; y[19] =-1.0; efl[19]=1;
	    x[20] = 1.0; y[20] = 0.0; efl[20]=1;
	    x[21] = 0.5*(x[18]+x[19]); y[21] = 0.5*(y[18]+y[19]); efl[21]=0;
	    x[22] = 0.5*(x[19]+x[20]); y[22] = 0.5*(y[19]+y[20]); efl[22]=1;
	    x[23] = 0.5*(x[20]+x[18]); y[23] = 0.5*(y[20]+y[18]); efl[23]=0;
	}

	if(shape_flag == 5)
	{
	    x[0] = 0.0;y[0] = 0.0; efl[0]=0;  /* first element */
	    x[1] = 1.0;y[1] = 0.0; efl[1]=1;
	    x[2] = 1.0;y[2] = 1.0; efl[2]=1;
	    x[3] = 0.5*(x[0]+x[1]);y[3] = 0.5*(y[0]+y[1]);efl[3]=0;
	    x[4] = 0.5*(x[1]+x[2]);y[4] = 0.5*(y[1]+y[2]);efl[4]=1;
	    x[5] = 0.5*(x[2]+x[0]);y[5] = 0.5*(y[2]+y[0]);efl[5]=0;

	    x[6] = 0.0;y[6] = 0.0; efl[6]=0;  /* second element */
	    x[7] = 1.0;y[7] = 1.0; efl[7]=1;
	    x[8] = 0.0;y[8] = 1.0; efl[8]=1;
	    x[9] = 0.5*(x[6]+x[7]);y[9] = 0.5*(y[6]+y[7]);efl[9]=0;
	    x[10] = 0.5*(x[7]+x[8]);y[10] = 0.5*(y[7]+y[8]);efl[10]=1;
	    x[11] = 0.5*(x[8]+x[6]);y[11] = 0.5*(y[8]+y[6]);efl[11]=0;

	    x[12] = 0.0;y[12] = 0.0; efl[12]=0;  /* third element */
	    x[13] = 0.0;y[13] = 1.0; efl[13]=1;
	    x[14] =-1.0;y[14] = 1.0; efl[14]=1;
	    x[15] = 0.5*(x[12]+x[13]);y[15] = 0.5*(y[12]+y[13]);efl[15]=0;
	    x[16] = 0.5*(x[13]+x[14]);y[16] = 0.5*(y[13]+y[14]);efl[16]=1;
	    x[17] = 0.5*(x[14]+x[12]);y[17] = 0.5*(y[14]+y[12]);efl[17]=0;

	    x[18] = 0.0;y[18] = 0.0; efl[18]=0;  /* fourth element */
	    x[19] =-1.0;y[19] = 1.0; efl[19]=1;
	    x[20] =-1.0;y[20] = 0.0; efl[20]=1;
	    x[21] = 0.5*(x[18]+x[19]);y[21] = 0.5*(y[18]+y[19]);efl[21]=0;
	    x[22] = 0.5*(x[19]+x[20]);y[22] = 0.5*(y[19]+y[20]);efl[22]=1;
	    x[23] = 0.5*(x[20]+x[18]);y[23] = 0.5*(y[20]+y[18]);efl[23]=0;

	    x[24] = 0.0;y[24] = 0.0; efl[24]=0;  /* fifth element */
	    x[25] =-1.0;y[25] = 0.0; efl[25]=1;
	    x[26] =-1.0;y[26] =-1.0; efl[26]=1;
	    x[27] = 0.5*(x[24]+x[25]);y[27] = 0.5*(y[24]+y[25]);efl[27]=0;
	    x[28] = 0.5*(x[25]+x[26]);y[28] = 0.5*(y[25]+y[26]);efl[28]=1;
	    x[29] = 0.5*(x[26]+x[24]);y[29] = 0.5*(y[26]+y[24]);efl[29]=0;

	    x[30] = 0.0;y[30] = 0.0; efl[30]=0;  /* sixth element */
	    x[31] =-1.0;y[31] =-1.0; efl[31]=1;
	    x[32] = 0.0;y[32] =-1.0; efl[32]=1;
	    x[33] = 0.5*(x[30]+x[31]);y[33] = 0.5*(y[30]+y[31]);efl[33]=0;
	    x[34] = 0.5*(x[31]+x[32]);y[34] = 0.5*(y[31]+y[32]);efl[34]=1;
	    x[35] = 0.5*(x[32]+x[30]);y[35] = 0.5*(y[32]+y[30]);efl[35]=0;

	    x[36] = 0.0;y[36] = 0.0; efl[36]=0;  /* seventh element */
	    x[37] = 0.0;y[37] =-1.0; efl[37]=1;
	    x[38] = 1.0;y[38] =-1.0; efl[38]=1;
	    x[39] = 0.5*(x[36]+x[37]);y[39] = 0.5*(y[36]+y[37]);efl[39]=0;
	    x[40] = 0.5*(x[37]+x[38]);y[40] = 0.5*(y[37]+y[38]);efl[40]=1;
	    x[41] = 0.5*(x[38]+x[36]);y[41] = 0.5*(y[38]+y[36]);efl[41]=0;

	    x[42] = 0.0;y[42] = 0.0; efl[42]=0;  /* eighth element */
	    x[43] = 1.0;y[43] =-1.0; efl[43]=1;
	    x[44] = 1.0;y[44] = 0.0; efl[44]=1;
	    x[45] = 0.5*(x[42]+x[43]);y[45] = 0.5*(y[42]+y[43]);efl[45]=0;
	    x[46] = 0.5*(x[43]+x[44]);y[46] = 0.5*(y[43]+y[44]);efl[46]=1;
	    x[47] = 0.5*(x[44]+x[42]);y[47] = 0.5*(y[44]+y[42]);efl[47]=0;

	}

/*
	dum = 0;
	printf( "numel    ndiv   circle flag  radius\n");
	printf( "%4d %4d %4d %16.8e\n", ne, ndiv, circle_flag, a);
	printf( "element   x   y   efl\n");
	for( i = 0; i < ne; ++i )
	{
	    printf( "%d\n", i);
	    for( j = 0; j < npel; ++j )
	    {
		printf( "   %16.8e %16.8e %4d", x[dum], y[dum], efl[dum]);
		printf( "\n");
		++dum;
	    }
	    if(npel == 3) dum += 3;
	}
*/

	check = trgl6( x, y, p, efl, c, ndiv, a);
	if(!check) printf( " Problems with trgl3_sqr \n");

	numnp = ng;
	numel = ne;

	check = slffeawrite( p, c, npel, nsd);
	if(!check) printf( " Problems with slffeawrite \n");

	return 1;

}


